#!/usr/bin/perl
#
# file: Sudoku.pm
#
# (c) Copyright J.Fossy Weinzinger 2005
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Fossy: 2005.11.03
# Fossy: 2005.11.14
#

use strict;
use warnings;

package Sudoku;

#----------------------------------------------------------------------------
my @lookupBits = 
  ( 1 <<  0,
    1 <<  1,
    1 <<  2,
    1 <<  3,
    1 <<  4,
    1 <<  5,
    1 <<  6,
    1 <<  7,
    1 <<  8,
  );

use constant ALLBITS => 0b1_1111_1111;
  
#----------------------------------------------------------------------------
sub Count1bits($)
{
  my $bits = shift;
  my $cnt  = 0;

  #
  # count the number of 1-bits in $bits
  #
  for ( ;
        $bits;
        $bits >>= 1
      )
    {
      $cnt++ if $bits & 1;
    }
    
  return $cnt;
}

my @lookupCard = map { Count1bits($_) } (0 .. ALLBITS);


#----------------------------------------------------------------------------
sub ReverseSet($)
#
# return first element (number) of set
#
{
  my $bits = shift;
  
  for ( my $i = 0;
        $bits;
        $i++, $bits >>= 1
      )
    {
      return $i if $bits & 1;  #============================================>
    }
    
  return undef;
}


#----------------------------------------------------------------------------
sub Index
{
  return 9 * $_[0] + $_[1];
}


#----------------------------------------------------------------------------
my @square;     # the sudoku square row1:0..8, row2:9..17, .. row9:72..80
my $changed;    # which indexes has changed ?
my $changedNow; # which indexes change now
my @results;
my $maxResults = 1e10;
my $maxDeep    = 0;
my $reduced;
my @measure;


#----------------------------------------------------------------------------
sub Read(@)
#
# convert input of numbers into sets
#
{
  for (my $idx = 0;
        @_;
        $idx++
      )
    {
      my $element = shift;
      
      if ( defined($element) && '1' le $element && $element le '9' )
        { $square[$idx] = $lookupBits[$element - 1]; }
      else
        { $square[$idx] = ALLBITS; }
    }
  
  #
  # let say *all* elements has changed
  #
  $changed    = { map { $_ => $square[$_] } (0 .. 9 * 9 - 1) };
  $changedNow = { };
}


#----------------------------------------------------------------------------
sub Write()
#
# convert sets into numbers
#
{
  $maxResults--;

  my @newSquare = map { ReverseSet($_) + 1; } @square;
  
  push @results, \@newSquare;
}


#----------------------------------------------------------------------------
sub ReducePartB(@)
{
  #
  # @_ contains 9 indexes of @square
  # 
  my @set     = @_;
  my $reduced = 0;

  for my $i (0..8)
    {
      my $s = $square[$set[$i]];
      
      if ( $lookupCard[$s] == 1 )
        {
          #
          # the element $i has exactly one element
          #
          for my $k (map { $set[$_] } grep { $_ != $i } (0..8))
            {
              if ( my $canDel = $s & $square[$k] )
                {
                  $square[$k]       &= ~$canDel;
                  $changedNow->{$k} |=  $canDel;
                  $reduced++;
                }
            }
        }
    }
  
  return $reduced;
}


#----------------------------------------------------------------------------
sub ReducePartA($)
{
  my $proj    = shift;
  my $reduced = 0;
  
  for my $slow (0..8)
    {
      my @subset;	# the indexes of @square

      for my $fast (0..8)
        {
          push @subset, Index($proj->($slow, $fast));
        }

      #
      # reduce subset if some of it has changed last time
      #
      $reduced += ReducePartB(@subset) if grep { $changed->{$_} } @subset;
    }

  return $reduced;
}


#----------------------------------------------------------------------------
sub div3($)
{ int($_[0] / 3) }


#----------------------------------------------------------------------------
sub mod3($)
{ $_[0] % 3 }


#----------------------------------------------------------------------------
sub Reduce()
{
  my $result = ReducePartA(sub { @_ })
             + ReducePartA(sub { reverse @_ })
             + ReducePartA(sub { my ($a, $b) = @_; (3*div3($a)+div3($b), 3*mod3($a)+mod3($b)) })
             ;
     
  $changed    = $changedNow;
  $changedNow = { };
  
  $reduced += $result;
  
  return $result;
}


#----------------------------------------------------------------------------
sub GiveBack($)
{
  my $giveback = shift;

  while ( my ($idx, $set) = each %$giveback )
    {
      $square[$idx] |= $set;
    }
}


#----------------------------------------------------------------------------
sub MergeChanged($$)
{
  my ($a, $b) = @_;

  while ( my ($idx, $set) = each %$b )
    {
      $a->{$idx} |= $set;
    }

  return $a;
}


#----------------------------------------------------------------------------
sub Backtracking($$);
sub Backtracking($$)
{
  my ($deep, $giveback) = @_;
  my $m = $reduced ;###== 0 ? 0. : 1. / $reduced;
  
  $measure[$deep] = $m if !defined($measure[$deep]) || $m > $measure[$deep];
  
  $maxDeep = $deep if $deep > $maxDeep;
  
  $deep++;
  
  if ( $maxResults <= 0 )
    {
      GiveBack($giveback);
      return;  #============================================================>
    }
  
  my @cnt = map { [] } (0..9);
  push @{$cnt[$lookupCard[$square[$_]]]}, $_ for (0 .. 9 * 9 - 1);

  #
  # @$cnt[0] hold all indexes of empty sets
  # @$cnt[1] hold all indexes of sets with exactly one elemnet
  # @$cnt[2] hold all indexes of sets with exactly two elemnets
  # ...
  #
  if ( @{$cnt[1]} == 9 * 9 )
    {
      #
      # found a result
      #
      Write();
      GiveBack($giveback);
      return;  #============================================================>
    }

  if ( @{$cnt[0]} > 0 )
    {
      # ups, there are empty sets
      GiveBack($giveback);
      return;  #============================================================>
    }

  #
  # find a element with minimum "open" elements
  #
  for (2..9)
    {
      if ( @{$cnt[$_]} > 0 )
        {
          my $idx = $cnt[$_]->[int(rand @{$cnt[$_]})];

          for ( my ($bits, $i) = ($square[$idx], 0);
                $bits;
                $bits >>= 1, $i++
              )
            {
              if ( $bits & 1 )
                {
                  my $chg = { $idx => $square[$idx] & ~$lookupBits[$i] };
                  
                  %$changed     = %$chg;
                  $square[$idx] = $lookupBits[$i];
                  $reduced      = 0;
                  
                  MergeChanged($chg, $changed) while Reduce();

                  Backtracking($deep, $chg);
                }
            }
            
          last;  #==========================================================v
        }
    }

  GiveBack($giveback);
}


#----------------------------------------------------------------------------
sub Solve($@)
{
  $maxResults = shift;
  $maxDeep    = 0;
  $reduced    = 0;
  @measure    = ();
  @results    = ();
  
  Read(@_);
  
  1 while Reduce();
  Backtracking(0, {});

  return @results;
}


#----------------------------------------------------------------------------
sub MaxDeep()
{
  return $maxDeep;
}


#----------------------------------------------------------------------------
sub Measure()
{
  my $measure = 0;
  
  $measure += $_ for ( @measure );
  
  return $measure + 100*$maxDeep;
}


#----------------------------------------------------------------------------
sub Class()
{
  my $measure = Measure;
  
  return 'easy'   if $measure <=  500;
  return 'medium' if $measure <=  800;
  return 'hard'   if $measure <= 1280;

  return 'daemonic';
}


#----------------------------------------------------------------------------
1;